﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using Microsoft.Examples.LtxOpenXml;
using DocumentFormat.OpenXml.Packaging;

class ContentItem
{
    public string Link { get; set; }
    public string Title { get; set; }
    public string Summary { get; set; }
    public string[] Authors { get; set; }
    public string[] Keywords { get; set; }
    public string ArticlePagesEstimate { get; set; }
    public string Code { get; set; }
    public string TechnicalLevel { get; set; }
    public string MediaType { get; set; }
    public string AuthorDate { get; set; }
}

class Program
{
#if true
    static string KeywordBaseUri = "open-xml-content-by-keyword.htm";
    static string AuthorBaseUri = "open-xml-content-by-author.htm";
#else
    static string KeywordBaseUri = "http://openxmldeveloper.org/wiki/w/wiki/open-xml-content-by-keyword.aspx";
    static string AuthorBaseUri = "http://openxmldeveloper.org/wiki/w/wiki/open-xml-content-by-author.aspx";
#endif

    // {0} Cell text
    static string TdTemplate =
@"<td valign=top style='border:solid windowtext 1.0pt;padding:0in 5.4pt 0in 5.4pt'>
<p style=''>{0}</p>
</td>
";

    // {0} Cell header
    static string ThTemplate =
@"<td valign=top style='border:solid windowtext 1.0pt;padding:0in 5.4pt 0in 5.4pt;font-weight: bold;'>
<p style=''>{0}</p>
</td>
";

    // {0} Table cells
    static string TrTemplate =
@"<tr style=''>
{0}
</tr>
";

    // {0} Table rows
    static string TableTemplate =
@"<table border='1' cellspacing='0' cellpadding='0'
 style='border-collapse:collapse;border:none;'>
{0}
</table>
";

#if true
    // {0} Table
    static string HtmlTemplate =
@"<html>
<head/>
<body>
{0}
</body>
</html>
";
#else
    // {0} Table
    static string HtmlTemplate =
@"{0}";
#endif

    static string[] SplitCommaSeparatedString(string unsplit)
    {
        if (unsplit == null)
            return null;
        string[] ss = unsplit.Split(',').Select(s => s.Trim()).ToArray();
        return ss;
    }

    static string KeyWordsToLinks(string[] keywords)
    {
        string links = keywords
            .Select(kw => KeywordToLink(kw) + "<br/>")
            .StringConcatenate();
        return links;
    }

    static string AuthorsToLinks(string[] authors)
    {
        if (authors == null)
            return "";
        string links = authors
            .Select(a => String.Format("<a href='{0}#{1}'>{2}</a><br/>", AuthorBaseUri, KeywordToId(a), a))
            .StringConcatenate();
        return links;
    }

    static string KeywordToId(string kw)
    {
        return kw.ToLower().Replace(' ', '-');
    }

    static string GenerateKeywordTable(List<ContentItem> contentList, string keyword)
    {
        string tableTitle = string.Format("<h1 id='{0}'>", KeywordToId(keyword)) + keyword + " Topics</h1>";
        string[] tableHeaders = new string[] {
            "Link",
            "Summary",
            "Author",
            "Keywords",
            "",
        };
        string table = String.Format(TableTemplate,
            String.Format(TrTemplate,
                tableHeaders.Select(th => String.Format(ThTemplate, th)).StringConcatenate())
            +
            contentList
                .Where(i => i.Keywords.Contains(keyword))
                .OrderBy(i => i.TechnicalLevel)
                .Select(i =>
                {
                    string s = String.Format(TrTemplate,
                        String.Format(TdTemplate,
                            String.Format("<a href='{0}' target='_blank'>{1}</a>",
                                i.Link, i.Title)) +
                        String.Format(TdTemplate, i.Summary) +
                        string.Format(TdTemplate, AuthorsToLinks(i.Authors)) +
                        string.Format(TdTemplate, KeyWordsToLinks(i.Keywords)) +
                        string.Format(TdTemplate, String.Format("<span style='color:#2b91af;'>Level:</span>&nbsp;{0}<br/><span style='color:#2b91af;'>Pages:</span>&nbsp;{1}<br/><span style='color:#2b91af;'>Code:</span>&nbsp;{2}<br/><span style='color:#2b91af;'>Media:</span>&nbsp;{3}",
                            i.TechnicalLevel, i.ArticlePagesEstimate, i.Code, i.MediaType)));

                    return s;
                }
                ).StringConcatenate()
        );
        return tableTitle + table;
    }

    static string GenerateAuthorTable(List<ContentItem> contentList, string author)
    {
        string tableTitle = string.Format("<h1 id='{0}'>", KeywordToId(author)) + "Topics by " + author + "</h1>";
        string[] tableHeaders = new string[] {
            "Link",
            "Summary",
            "Author",
            "Keywords",
            "",
        };
        string table = String.Format(TableTemplate,
            String.Format(TrTemplate,
                tableHeaders.Select(th => String.Format(ThTemplate, th)).StringConcatenate())
            +
            contentList
                .Where(i => i.Authors != null && i.Authors.Contains(author))
                .OrderBy(i => i.TechnicalLevel)
                .Select(i =>
                {
                    string s = String.Format(TrTemplate,
                        String.Format(TdTemplate,
                            String.Format("<a href='{0}' target='_blank'>{1}</a>",
                                i.Link, i.Title)) +
                        String.Format(TdTemplate, i.Summary) +
                        string.Format(TdTemplate, AuthorsToLinks(i.Authors)) +
                        string.Format(TdTemplate, KeyWordsToLinks(i.Keywords)) +
                        string.Format(TdTemplate, String.Format("<span style='color:#2b91af;'>Level:</span>&nbsp;{0}<br/><span style='color:#2b91af;'>Pages:</span>&nbsp;{1}<br/><span style='color:#2b91af;'>Code:</span>&nbsp;{2}<br/><span style='color:#2b91af;'>Media:</span>&nbsp;{3}",
                            i.TechnicalLevel, i.ArticlePagesEstimate, i.Code, i.MediaType)));

                    return s;
                }
                ).StringConcatenate()
        );
        return tableTitle + table;
    }

    static string GenerateKeywordTables(List<ContentItem> contentList)
    {
        var uniqueKeywords = contentList.SelectMany(c => c.Keywords).OrderBy(c => c).Distinct();
        return uniqueKeywords.Select(k => GenerateKeywordTable(contentList, k)).StringConcatenate();
    }

    static string GenerateAuthorTables(List<ContentItem> contentList)
    {
        var authors = contentList.Where(a => a.Authors != null).SelectMany(c => c.Authors).OrderBy(c => c).Distinct();
        return authors.Select(k => GenerateAuthorTable(contentList, k)).StringConcatenate();
    }

    static string KeywordToLink(string kw)
    {
        var s = string.Format("<a href='{0}#{1}'>{2}</a>", KeywordBaseUri, KeywordToId(kw), kw);
        return s;
    }

    static string GenerateTOC(List<ContentItem> contentList)
    {
        string[] keywords = contentList.SelectMany(i => i.Keywords).OrderBy(k => k).Distinct().ToArray();

        string rows = keywords.Select((k, i) => new
        {
            kw = k,
            chunk = i / 5,
        })
        .GroupBy(k => k.chunk)
        .Select(g =>
            String.Format("<tr style=''>{0}</tr>", g.Select(k =>
                String.Format("<td style='border:solid windowtext 1.0pt;padding:3.5pt 5.4pt 4.5pt 5.4pt'>{0}</td>", KeywordToLink(k.kw))).StringConcatenate())).StringConcatenate();
        string table = String.Format("<table border='1' cellspacing='0' cellpadding='0' style='border-collapse:collapse;border:none;'>{0}</table>", rows);
        return table;
    }

    public static void TransformToHtml(string filename)
    {
        using (SpreadsheetDocument spreadsheet = SpreadsheetDocument.Open(filename, false))
        {
            //Console.WriteLine("Content List");
            //Console.WriteLine("==================");

            //// list all of the columns in the Inventory table
            //Console.WriteLine("Table: Content List");
            //foreach (var c in spreadsheet.Table("ContentList").TableColumns())
            //    Console.WriteLine("  {0}", c.Name);
            //Console.WriteLine();
            //Environment.Exit(0);

            //Content List
            //==================
            //Table: Content List
            //  SrNo
            //  Link
            //  Title
            //  Summary
            //  Author
            //  Keyword
            //  Article Pages (Estimate)
            //  Code
            //  Technical_x000a_Level
            //  Media Type
            //  Author Date
            //  Question
            //  Note

            var query = spreadsheet.Table("ContentList").TableRows();

            var contentList = query.Select(r => new ContentItem()
            {
                Link = r["Link"].Value,
                Title = r["Title"].Value,
                Summary = r["Summary"].Value,
                Authors = SplitCommaSeparatedString(r["Author"].Value),
                Keywords = SplitCommaSeparatedString(r["Keyword"].Value),
                ArticlePagesEstimate = r["Article Pages (Estimate)"].Value == null ? "" : r["Article Pages (Estimate)"].Value,
                Code = r["Code"].Value == null ? "" : r["Code"].Value,
                TechnicalLevel = r["Technical_x000a_Level"].Value == null ? "" : r["Technical_x000a_Level"].Value,
                MediaType = r["Media Type"].Value == null ? "" : r["Media Type"].Value,
                AuthorDate = r["Author Date"].Value == null ? "" : r["Author Date"].Value,
            }).Skip(1).ToList();

            string page = string.Format(HtmlTemplate,
                GenerateTOC(contentList) +
                GenerateKeywordTables(contentList));

            File.WriteAllText("open-xml-content-by-keyword.htm", page);

            string page2 = string.Format(HtmlTemplate,
                GenerateAuthorTables(contentList));
            File.WriteAllText("open-xml-content-by-author.htm", page2);
        }
    }

    public static void Main(string[] args)
    {
        TransformToHtml("../../OpenXmlContent.xlsx");
    }
}
